//----------------------------------------------------------------------------------------------
// BSP : Seeed nRF52 Borads 1.1.1
// Board : Seeed nRF52 Borads / Seeed XIAO nRF52840 Sense 
//----------------------------------------------------------------------------------------------

// 2023/07/26

#include <Wire.h>
#include "SparkFunBME280.h"
#include <Arduino.h>
#include <Adafruit_TinyUSB.h>   // For Serial.print()
#include <bluefruit.h>          // /Arduino15/packages/Seeeduino/hardware/nrf52/1.1.1/libraries/Bluefruit52Lib/src
#include <Adafruit_SPIFlash.h>  // Need to be deleted /Documents/Arduino/libraries/SdFat

#define HICHG         22        // Charge current　Hi:50mA Lo:100mA       D22 (13u)
#define CHG           23        // Charge LED Lo:ON                       D23 (17u)
#define VBAT_ENABLE   14        // Battery voltage read enable  Lo:enable D14 (14u)   
#define VBAT_READ     32        // Battery voltage read pin               D32 (31u)
#define FILE_NAME "log.csv"     // For the datalogging file:
//#define SLEEP_TIME    60        // Sleep time [sec]
#define SLEEP_TIME    2         // for TEST
#define Vbatt_LOW1    3.7       // Vbatt Low limit voltage
#define Vbatt_LOW2    3.5       // Vbatt critical Low limit voltage
//#define WAKEUP_COUNT  60        // Measure and data save by WAKEUP_COUNT * SLEEP_TIME
#define WAKEUP_COUNT  2         // for TEST

// Built from the P25Q16H datasheet.
// https://gitlab.com/arduino5184213/seeed_xiao_nrf52840/flash_speedtest/-/tree/master
SPIFlash_Device_t const P25Q16H {
  .total_size = (1UL << 21), // 2MiB
  .start_up_time_us = 10000, // Don't know where to find that value
  
  .manufacturer_id = 0x85,
  .memory_type = 0x60,
  .capacity = 0x15,

  .max_clock_speed_mhz = 55,
  .quad_enable_bit_mask = 0x02, // Datasheet p. 27
  .has_sector_protection = 1,   // Datasheet p. 27
  .supports_fast_read = 1,      // Datasheet p. 29
  .supports_qspi = 1,           // Obviously
  .supports_qspi_writes = 1,    // Datasheet p. 41
  .write_status_register_split = 1, // Datasheet p. 28
  .single_status_byte = 0,      // 2 bytes
  .is_fram = 0,                 // Flash Memory
};

Adafruit_FlashTransport_QSPI flashTransport;
Adafruit_SPIFlash flash(&flashTransport);
FatVolume fatfs;  // file system object from SdFat
BME280 mySensor;

uint32_t mS;                  // Measurement time
float Vbatt;                  // Battery voltage [V]
float Humidity;               // Measurement data
float Pressure;
float Temperature;
uint8_t Count = 0;            // wake up counter

//*********************************************************************************************
void setup()
{
  // Serial port initialization
  Serial.begin(115200);
//  while (!Serial);
  delay(2000);
  
  // Pin initialization
  pinMode(VBAT_READ, INPUT); 
  pinMode(HICHG, OUTPUT);
  pinMode(VBAT_ENABLE, OUTPUT);   
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);
  pinMode(LED_BLUE, OUTPUT);

  digitalWrite(HICHG, LOW);         // Charge current 100mA
  digitalWrite(VBAT_ENABLE, LOW);   // Battery voltage read enable   
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  digitalWrite(LED_BLUE, HIGH);

  // ADC initialization　See wireing_analog_nRF52.c:73
  analogReference(AR_DEFAULT);      // Default 0.6V*6=3.6V
  analogReadResolution(12);         // wireing_analog_nRF52.c:39
  analogSampleTime(10);             // default 3uS　3,5,10,15,20,40
    
  // BME280 Initialization
  Wire.begin();
  Wire.setClock(400000);            // Increase to fast I2C speed!

  mySensor.setI2CAddress(0x76);
  if (mySensor.beginI2C() == false) // Begin communication over I2C
  {
    Serial.println("The sensor did not respond. Please check wiring.");
    while(1){
      digitalWrite(LED_RED, LOW);
      delay(100);
      digitalWrite(LED_RED, HIGH);
      delay(100);
    }
  }
  mySensor.setFilter(0); // 0 to 4 is valid. Filter coefficient. See 3.4.4
  mySensor.setStandbyTime(0); //0 to 7 valid. Time between readings. See table 27.
  mySensor.setTempOverSample(1); //0 to 16 are valid. 0 disables temp sensing. See table 24.
  mySensor.setPressureOverSample(1); //0 to 16 are valid. 0 disables pressure sensing. See table 23.
  mySensor.setHumidityOverSample(1); //0 to 16 are valid. 0 disables humidity sensing. See table 19.   
  Serial.println("BME280 Initialization end");

  // Enable DC-DC converter Mode
  NRF_POWER->DCDCEN = 1;            // Enable DC/DC converter for REG1 stage
  
  Serial.println("End of initialization");
}

//************************************************************************************************************
void loop()
{
  // wake up     
    digitalWrite(LED_GREEN, LOW);
    
    Count = Count + 1;
    
    // Battery voltage averaging
    int Vadc = 0;
    for (int i = 0; i < 16; i++) {
      Vadc = Vadc + analogRead(VBAT_READ);
    }
    Vadc = Vadc / 16;
    Vbatt = 2.961 * 3.6 * Vadc / 4096;   // Vref=3.6  ratio=2.961

    // Battery Voltage check
    if (Vbatt < Vbatt_LOW2) {
      digitalWrite(LED_GREEN, HIGH);    // GREEN : Normal voltage
      digitalWrite(LED_RED, LOW);       // RED : critical Low voltage      
    } else if (Vbatt < Vbatt_LOW1) {
      digitalWrite(LED_GREEN, HIGH);
      digitalWrite(LED_BLUE, LOW);      // BLUE : Low voltage    
    }    

    // Measure and data save
    if (Count >= WAKEUP_COUNT)
    {     
      digitalWrite(LED_GREEN, LOW);
      digitalWrite(LED_RED, HIGH);
      digitalWrite(LED_BLUE, HIGH);
      
      Count = 0;
       
      // BME280 start measurement
      mySensor.setMode(MODE_FORCED);
      while(mySensor.isMeasuring() == false) ; //Wait for sensor to start measurment

      // Measurement time
      mS = millis();

      // Initialize Flash library and check its chip ID.
      Serial.println("Initialize flash library and check its chip ID");
      if (!flash.begin(&P25Q16H, 1)) {
        Serial.println("Error, failed to initialize flash chip!");
        while(1) {
          digitalWrite(LED_RED, LOW);
          delay(200);
          digitalWrite(LED_RED, HIGH);
          delay(200);
        }
      }
      Serial.print("Flash chip JEDEC ID: 0x"); Serial.println(flash.getJEDECID(), HEX);

      // mount the filesystem
      Serial.println("mount the filesystem");
      if (!fatfs.begin(&flash)) {
        Serial.println("Error, failed to mount newly formatted filesystem!");
        Serial.println("Was the flash chip formatted with the fatfs_format example?");
        while(1) {
          digitalWrite(LED_RED, LOW);
          delay(300);
          digitalWrite(LED_RED, HIGH);
          delay(300);
        }
      }
      Serial.println("Mounted filesystem!");

      // log file open
      Serial.println("file open");
      File32 dataFile = fatfs.open(FILE_NAME, FILE_WRITE);
      Serial.println("file opened");
      if (dataFile)
      {      
        // save sensor data
        Serial.println("BME280 data read");
        while(mySensor.isMeasuring() == true) ; // While sensor completes the reading  7mS
        Humidity = mySensor.readFloatHumidity();
        Pressure = mySensor.readFloatPressure();
        Temperature = mySensor.readTempC();

        dataFile.print(mS);
        dataFile.print(",");
        dataFile.print(Humidity, 3);
        dataFile.print(",");
        dataFile.print(Pressure, 3);
        dataFile.print(",");
        dataFile.print(Temperature, 3);
        dataFile.print(",");      
        dataFile.println(Vbatt, 3);

        Serial.print(mS);
        Serial.print(",  ");
        Serial.print(Humidity, 3);
        Serial.print(",  ");
        Serial.print(Pressure, 3);
        Serial.print(",  ");
        Serial.print(Temperature, 3);
        Serial.print(",  ");      
        Serial.println(Vbatt, 3);

        // file close, fatfs end, flash end
        Serial.println("file close, fatfs end, flash end");
        dataFile.close();
        fatfs.end();
        flash.end();
      }
      else
      {
        Serial.println("Failed to open data file for writing!");
        while(1) {
          digitalWrite(LED_RED, LOW);
          delay(400);
          digitalWrite(LED_RED, HIGH);
          delay(400);
        }
      }
     Serial.println();
   }

  digitalWrite(LED_GREEN, HIGH);
  digitalWrite(LED_BLUE, HIGH);
  digitalWrite(LED_RED, HIGH);
       
  delay(SLEEP_TIME * 1000);
}
